// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  CatalogDomainInfo.java
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// wi8754       WXS 8.5   03/23/2012   cdjohnson new part   
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid;

import com.ibm.websphere.objectgrid.security.config.ClientSecurityConfiguration;

/**
 * Identifies the configuration attributes of a catalog service domain.
 * @since 8.5, XC10
 * @see CatalogDomainManager
 */
public interface CatalogDomainInfo {
    
    /**
     * Retrieve the identifier of the domain as specified by the catalog service domain.<p>  
     * 
     * Note:  This is different than the name of the domain which is specified when starting the
     * catalog services.
     * @return the identifier of the domain.
     */
    public String getDomainId();
    
    /**
     * Retrieve the catalog server endpoints used to connect a client to the remote
     * catalog service domain.<p>
     * 
     * The catalog service endpoints are used with the 
     * {@link ObjectGridManager#connect(String, com.ibm.websphere.objectgrid.security.config.ClientSecurityConfiguration, java.net.URL)}
     * method to connect to a catalog service domain.
     * @return the catalog service endpoints.
     */
    public String getClientCatalogServerEndpoints();
    
    /**
     * Retrieve the ClientSecurityConfiguration for the domain.<p>
     * 
     * The ClientSecurityConfiguration are used with the 
     * {@link ObjectGridManager#connect(String, com.ibm.websphere.objectgrid.security.config.ClientSecurityConfiguration, java.net.URL)}
     * method to connect to a catalog service domain.
     * @return the ClientSecurityConfiguration or null if security is not configured.
     */
    public ClientSecurityConfiguration getClientSecurityConfiguration();
}
