// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2008
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ClientClusterContext.java
//
// Source File Description: Reference JavaDoc
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// xxxxxx       WASX.XD   09/02/2005   rajeshk    Adding support for remote client
// xxxxxx       WASX.XD   12/12/2005   haowang    add ripple update and thread safe
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid;

import java.net.URL;

import com.ibm.websphere.objectgrid.client.ClientProperties;

/**
 * This interface is a context to represent which cluster/domain the client connected to
 * using one of the <code>ObjectGridManager.connect</code> methods.  An instance
 * of this interface is used to retrieve client <code>ObjectGrid</code> instances
 * and for performing admin operations against an ObjectGrid cluster/domain or its servers.
 *
 * @since WAS XD 6.0.1, XC10
 *
 * @see ObjectGridManager
 */
public interface ClientClusterContext {

    /**
     * Gets the name of the domain to which the client is connected
     *
     * @return the name of the domain this context is connected to
     */
    String getClusterName();

	/**
	 * Retrieve the ClientProperties object for this ClientClusterContext for the
	 * specified ObjectGrid name.  A ClientProperties is scoped to this
	 * ClientClusterContext and a single ObjectGrid.
	 *
	 * @param objectGridName the name of ObjectGrid
	 *
	 * @return the ClientProperties instance for this ObjectGrid.
	 * @since WAS XD 6.1.0.3
	 */
	public ClientProperties getClientProperties(String objectGridName);

	/**
     * Sets the ClientProperties properties for the selected ObjectGrid using the
     * specified client properies file.<p>
     *
     * To further adjust the client properties, call the {@link #getClientProperties(String)} method.
     *
     * @param objectGridName the name of ObjectGrid to apply the ClientProperties to.
     * @param url the URL where the client properties file can be located.
     * @since WAS XD 6.1.0.3
     */
    public void setClientProperties(String objectGridName, URL url);
}
