// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ObjectGridException.java
//
// Source File Description:
//
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   01/23/2005   rajeshk   new code drop
// bug 10193    6.1 fix3  10/16/2007   jhanders  Update exception constructors
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid;

/**
 * Base exception class for all checked exceptions thrown by the ObjectGrid product.
 *
 * @since WAS XD 6.0, XC10
 */
public class ObjectGridException extends Exception implements IObjectGridException
{
    /**
     * The <code>serialVersionUID</code> for this class.
     */
    private static final long serialVersionUID = 5441454934720178237L;

    /**
     * The throwable that caused this ObjectGridException to get thrown, or null if this
     * throwable was not caused by another throwable, or if the causative
     * throwable is unknown.  If this field is equal to this ObjectGridException itself,
     * it indicates that the cause of this ObjectGridException has not yet been
     * initialized.
     */
    private Throwable ivCause = this;

    /**
     * Constructs a new ObjectGridException with <code>null</code> as its detail
     * message. The cause is not initialized, and may subsequently be initialized
     * by a call to the <code>initCause</code> method.
     *
     * @see #initCause(Throwable)
     */
    public ObjectGridException()
    {
        super();
    }

    /**
     * Constructs a new ObjectGridException with the specified detail message. The
     * cause is not initialized, and may subsequently be initialized by a call to
     * the <code>initCause</code> method.
     *
     * @param message
     *           the detail message. The detail message is saved for later
     *           retrieval by the <code>getMessage</code> method.
     *
     * @see #initCause(Throwable)
     * @see #getMessage()
     */
    public ObjectGridException(String message)
    {
        super(message);
    }

    /**
     * Constructs a new ObjectGridException with a specified cause.
     * The cause and a detail message of <code>(cause==null ? null : cause.toString())</code>
     * is used (which typically contains the class and detail message of <code>cause</code>).
     * This constructor is useful for ObjectGridExceptions that are little more
     * than wrappers for other throwables.
     *
     * @param cause is the exception that caused this exception to be thrown,
     *        which is saved for later retrieval by the <code>getCause()</code> method.
     *        A <code>null</code> value is permitted and indicates that the cause
     *        is nonexistent or is unknown.
     *
     * @see #getCause()
     */
    public ObjectGridException(Throwable cause)
    {
        super(cause);
        ivCause = cause;
    }

    /**
     * Constructs a new ObjectGridException with the specified detail message and
     * cause.
     * <p>
     * Note that the detail message associated with <code>cause</code> is
     * <i>not</i> automatically incorporated in this ObjectGridException's detail
     * message.
     *
     * @param message
     *           the detail message (which is saved for later retrieval by the
     *           <code>getMessage</code> method).
     * @param cause
     *           the cause (which is saved for later retrieval by the
     *           <code>getCause</code> method). (A<code>null</code> value is
     *           permitted, and indicates that the cause is nonexistent or
     *           unknown).
     *
     * @see #getCause()
     * @see #getMessage()
     */
    public ObjectGridException(String message, Throwable cause)
    {
        super(message, cause);
        ivCause = cause;
    }

    /**
     * Returns the cause of this ObjectGridException or <code>null</code> if the
     * cause is nonexistent or unknown.  (The cause is the throwable that
     * caused this ObjectGridException to get thrown.)
     *
     * <p>This implementation returns the cause that was supplied via one of
     * the constructors requiring a <code>Throwable</code>, or that was set after
     * creation with the <code>initCause(Throwable)</code> method.  While it is
     * typically unnecessary to override this method, a subclass can override
     * it to return a cause set by some other means.  This is appropriate for
     * a "legacy chained throwable" that predates the addition of chained
     * exceptions to <code>Throwable</code>.  Note that it is <i>not</i>
     * necessary to override any of the <code>PrintStackTrace</code> methods,
     * all of which invoke the <code>getCause</code> method to determine the
     * cause of an ObjectGridException
     *
     * @return  the cause of this ObjectGridException or <code>null</code> if the
     *          cause is nonexistent or unknown.
     *
     * @see #ObjectGridException(String, Throwable)
     * @see #ObjectGridException(Throwable)
     * @see #initCause(Throwable)
     */
    public Throwable getCause() {
        return super.getCause();
    }

    /**
     * Initializes the <i>cause</i> of this ObjectGridException to the specified value.
     * (The cause is the throwable that caused this ObjectGridException to get thrown.)
     *
     * <p>This method can be called at most once.  It is generally called from
     * within the constructor, or immediately after creating the
     * ObjectGridException.  If this ObjectGridException was created
     * with <code>ObjectGridException(Throwable)</code> or
     * <code>ObjectGridException(String,Throwable)</code>, this method cannot be called
     * even once.
     *
     * @param  cause the cause (which is saved for later retrieval by the
     *         <code>getCause()</code> method).  (A <code>null</code> value is
     *         permitted, and indicates that the cause is nonexistent or
     *         unknown.)
     * @return a reference to this <code>ObjectGridException</code> instance.
     * @throws IllegalArgumentException if <code>cause</code> is this
     *         ObjectGridException.  (An ObjectGridException cannot be its own cause.)
     * @throws IllegalStateException if this ObjectGridException was
     *         created with <code>ObjectGridException(Throwable)</code> or
     *         <code>ObjectGridException(String,Throwable)</code>, or this method has already
     *         been called on this ObjectGridException.
     *
     * @see #ObjectGridException(String, Throwable)
     * @see #ObjectGridException(Throwable)
     * @see #getCause()
     */
    public synchronized Throwable initCause(Throwable cause) {
        super.initCause(cause);
        ivCause = cause;
        return this;
    }
}