// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2010
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  LRUEvictor.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   02/03/2005   sutter    new code drop
// 267383       WASX.XD   04/20/2005   jhanders  touch support
// XD60_M1_B.12 WASX.XD   04/26/2005   sutter    Make Eviction (LRU/LFU) processing work
// 272056       WASX.XD   05/03/2005   jhanders  combine changes() and evictees()
// XD60_M4_B.21 WASX.XD   05/23/2005   sutter    Client/server optimizations
// 278516       WASX.XD   05/24/2005   kjlaw     eliminate loop problem due to thread safety issues.
// 287285       WASX.XD   07/28/2005   kjlaw     make changes required as new BaseMapLogSequence updates.
// 287677       WASX.XD   06/29/2005   kjlaw     use new EvictorData interfaces.
// 294637.1     WASX.XD   08/04/2005   kjlaw     implement RollbackEvictor interface.
// XD601_M3_B.x WASX.XD   10/28/2005   kjlaw     add getNumberOfHeaps method.
// d328136      WASX.XD   12/08/2005   kjlaw     externalize EvictorData.
// d332063      WASX.XD   12/13/2005   kjlaw     destroy needs to use doPriv
// d335510      WASX.XD   01/04/2006   jitang    Use AccessController.doPrivileged
// d337865.1    WXD601.XD 02/12/2006   kjlaw     apply needs to handle incorrect order better
// d347038      WXD601.XD 02/14/2006   kjlaw     Need to maintain EvictorData even when map is not activated.
// d350934      WXD601.XD 02/28/2006   kjlaw     clear Queues on deactivate.
// bug 11228    6.1.0.2   02/07/2008   jhanders  Update Evictor to not output IllegalArgumentException
// bug 11502    6.1.0.3   03/14/2008   cheng1    enable memory based eviction.
// bug  12770   6.1.0.3   07/03/2008   cheng1    allow unlimited size indicates by maxSize=0 when useMaxMemory=true.
// bug 18905    7.1       04/23/2010   bsandman  added disk delegator abstraction
// b18690       7.1       05/06/2010   jhanders  Update to not use reflection
// bug 18690    7.1       05/14/2010   jhanders  Code review comments for Evictor disk offload implementation
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins.builtins;

import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.plugins.CacheEntry;
import com.ibm.websphere.objectgrid.plugins.EvictionEventCallback;
import com.ibm.websphere.objectgrid.plugins.Evictor;
import com.ibm.websphere.objectgrid.plugins.EvictorData;
import com.ibm.websphere.objectgrid.plugins.LogElement;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.RollbackEvictor;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.websphere.objectgrid.server.ServerFactory;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.DiffMapValue;
import com.ibm.ws.objectgrid.ObjectGridImpl;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.Storage;
import com.ibm.ws.objectgrid.io.offheap.LRUOffHeapEvictionData;
import com.ibm.ws.objectgrid.io.offheap.LRUOffHeapEvictionDelegator;
import com.ibm.ws.objectgrid.io.offheap.ObjectGridHashTableOH;
import com.ibm.ws.objectgrid.io.offheap.OffHeapEvictionData;
import com.ibm.ws.objectgrid.io.offheap.OffHeapManager;
import com.ibm.ws.objectgrid.io.offheap.impl.XsOffHeapEvictionDataImpl;
import com.ibm.ws.objectgrid.keys.KeyType;
import com.ibm.ws.objectgrid.management.MemoryEvictionController;
import com.ibm.ws.objectgrid.management.MemoryEvictionControllerImpl;
import com.ibm.ws.objectgrid.management.MemoryManagement;
import com.ibm.ws.objectgrid.management.MemoryManagementImpl;
import com.ibm.ws.objectgrid.map.BaseMap;
import com.ibm.ws.objectgrid.map.SystemMap;
import com.ibm.ws.objectgrid.plugins.DiskLRUEvictorData;
import com.ibm.ws.objectgrid.plugins.EvictionDelegator;
import com.ibm.ws.objectgrid.plugins.LRUEvictionDelegator;
import com.ibm.ws.objectgrid.plugins.LRUEvictorData;
import com.ibm.ws.objectgrid.plugins.LRUEvictorDataTransformer;
import com.ibm.ws.objectgrid.plugins.MemoryEvictionTrigger;
import com.ibm.ws.objectgrid.plugins.SystemCacheEntry;
import com.ibm.ws.objectgrid.server.ServerPropertiesImpl;
import com.ibm.ws.objectgrid.util.DiskHelper;
import com.ibm.ws.objectgrid.util.DiskHelperFactory;
import com.ibm.ws.objectgrid.util.ObjectGridUtil;
import com.ibm.ws.objectgrid.util.Queue;
import com.ibm.ws.xs.NLSConstants;
import com.ibm.ws.xs.thread.ScheduledTask;
import com.ibm.ws.xs.thread.ThreadPool;
import com.ibm.ws.xs.thread.ThreadPoolManagerFactory;
import com.ibm.ws.xs.util.DiskOverflowMapIdentifier;
import com.ibm.ws.xs.util.XSUtilities;

/**
 * This evictor manages a BackingMap using a simple Least Recently Used (LRU) algorithm. It attempts to keep the
 * BackingMap at less than a certain number of entries.
 *
 * @since WAS XD 6.0, XC10
 *
 * @see Evictor
 */
@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
final public class LRUEvictor implements Evictor, RollbackEvictor, Runnable {
    private final static String CLASS_NAME = LRUEvictor.class.getName();

    private final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_EVICTOR_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * Default sleep time for evictor thread if the <code>setSleepTime(int)</code> method is not called.
     *
     * @see #setSleepTime(int)
     */
    public static final long DEFAULT_SLEEP_TIME = 15000L;

    /**
     * Default number of LRU queues to create if the <code>setNumberOfLRUQueues(int)</code> method is not called.
     *
     * @see #setNumberOfLRUQueues(int)
     */
    public static final int DEFAULT_NUMBER_OF_QUEUES = 51;

    private static final Method triggerEvictionMethod;

    static {
        try {
            triggerEvictionMethod = (Method) AccessController.doPrivileged(new PrivilegedExceptionAction() {

                public Object run() throws Exception {
                    Method m = LRUEvictor.class.getDeclaredMethod("triggerEviction", new Class[0]);
                    m.setAccessible(true);
                    return m;
                }

            });
        } catch (Exception e) {
            IllegalStateException ise = new IllegalStateException(
                    "Failed to initialize class due to following exception");
            ise.initCause(e);
            throw ise;
        }
    }

    /** The EvictionEventCallback instance */
    EvictionEventCallback ivCallback;

    BackingMap ivBackingMap;

    int ivNumberOfQueues = DEFAULT_NUMBER_OF_QUEUES;

    Queue ivQueues[];

    int ivMaxSize = 0;

    private long ivSleepTime = DEFAULT_SLEEP_TIME;

    private MemoryManagement memoryManagement;

    private MemoryEvictionTrigger listener;

    private boolean ivUseMemoryUsageThresholdEviction = false;

    private boolean ivUseGridCapThresholdEviction = false;

    private volatile boolean ivQuit = true;

    private boolean ivLogInvalidLogElement = true;

    private boolean ivDestroyed = false;

    private boolean ivActivated = false;

    private static boolean hasOutputUsageThresholdNotSupportedMsg = false;

    private String evictorId = " ";

    private ScheduledTask scheduledEviction;

    private Thread activeThread;

    private final ThreadPool threadPool = ThreadPoolManagerFactory.getThreadPoolManager().getThreadPool("LRUEvictor");

    private ClassLoader classLoader;

    private MemoryEvictionController memEvictionController;

    private boolean ivUseMemoryEvictionController = false;

    private boolean disk = false;

    private boolean offheap = false;

    private LRUEvictionDelegator delegator = null;

    //Keep track of heap size in memory when using DISK.
    private QueueCount[] queueCounts = null;

    private static class QueueCount {
        int count = 0;
    }

    // for NDO
    private String mapId;
    private String mapSetId;

    private int divisor = 0;

    /**
     * Creates a new LRUEvictor object with default values for the maximum size per LRU queue, the number of queues, and
     * sleep time between sweeps by the evictor thread.
     * <p>
     * The default values can be overridden by use of the <code>setMaxSize(int)</code>,
     * <code>setNumberOfLRUQueues(int)</code>, and <code>setSleepTime(int)</code> methods. If the
     * <code>setMaxSize</code> method is never called, the size of BackingMap is unlimited and no LRU data is kept.
     */
    public LRUEvictor() {
        super();
    }

    /**
     * @see Evictor#initialize(BackingMap, EvictionEventCallback)
     */
    public void initialize(BackingMap map, EvictionEventCallback callback) {
        evictorId = "LRUEvictorForMap-" + map.getName() + "_" + map.getPartitionId();
        classLoader = Thread.currentThread().getContextClassLoader();
        ivCallback = callback;
        ivBackingMap = map;

        // Get storage mode
        disk = ((BaseMap) ivBackingMap).getStorage() == Storage.DISK;
        offheap = ((BaseMap) ivBackingMap).getStorage() == Storage.OFF_HEAP;

        int numPartitions = map.getPartitionManager().getNumOfPartitions();
        if (numPartitions > 1 && numPartitions == ivNumberOfQueues) {
            divisor = ivNumberOfQueues;
        }

        // if disk skip Queue creation and memory initialization...will take out
        // all memory processing in flow
        if (disk || offheap) {

            ((BaseMap) ivBackingMap).setEvictorDataTransformer(LRUEvictorDataTransformer.instance);

            if (disk) {
                DiskHelper diskHelper = DiskHelperFactory.getDiskHelper();
                delegator = (LRUEvictionDelegator) diskHelper.createEvictionDelegator(EvictionDelegator.LRU_DERBY,
                        ivNumberOfQueues);
            } else /*offheap*/{
                delegator = new LRUOffHeapEvictionDelegator();
                String evictionTriggers = ivBackingMap.getEvictionTriggers();
                if (evictionTriggers != null) {
                    ivUseMemoryUsageThresholdEviction = ObjectGridUtil.isTriggerInEvictionTriggers(
                        BackingMap.EVICTIONTRIGGER_MEMORY_USAGE_THRESHOLD, evictionTriggers);
                    ivUseGridCapThresholdEviction=ObjectGridUtil.isTriggerInEvictionTriggers(Constants.EVICTIONTRIGGER_GRID_CAP, evictionTriggers);
                    // grid cap eviction will be triggered by NOF when an insert occurs which exceeds the cap.
                    // no need to periodically wake up and check.
                    if(ivUseGridCapThresholdEviction) setSleepTime(Integer.MAX_VALUE);
                }
            }

            try {
                delegator.initialize(map, evictorId);

            } catch (Throwable e) {
                FFDCFilter.processException(e, CLASS_NAME + ".init", "286", this);
                throw new RuntimeException(e);
            }

            //Using an array of Objects vs a primitive array here because a primitive array is a single
            //object and would need to be protected by a single monitor (memory barrier).  We use
            //multiple monitors in our bucket approach for evictors to improve concurrency.
            queueCounts = new QueueCount[ivNumberOfQueues];
            for (int i = 0; i < ivNumberOfQueues; i++) {
                queueCounts[i] = new QueueCount();
            }

            //There could be evictor data on the disk so we need to initialize the in-memory records
            try {
                int[] qcs = delegator.getQueueCounts();
                for (int i = 0; i < this.ivNumberOfQueues; i++) {
                    synchronized (queueCounts[i]) {
                        queueCounts[i].count = qcs[i];
                    }
                }
            } catch (Throwable e) {
                FFDCFilter.processException(e, CLASS_NAME + ".initalize", "252");
            }
        } else {
            // get BackingMap's trigger.
            String evictionTriggers = ivBackingMap.getEvictionTriggers();
            if (evictionTriggers != null) {
                // parse the evictionTriggers to see if memory eviction is
                // specified.
                ivUseMemoryUsageThresholdEviction = ObjectGridUtil.isTriggerInEvictionTriggers(
                        BackingMap.EVICTIONTRIGGER_MEMORY_USAGE_THRESHOLD, evictionTriggers);

                if (ivUseMemoryUsageThresholdEviction) {
                    ivUseMemoryEvictionController = false;
                } else {
                    ivUseMemoryEvictionController = ObjectGridUtil.isTriggerInEvictionTriggers(
                            MemoryEvictionController.MEMORY_USAGE_MONITOR_TRIGGER, evictionTriggers);
                }
                
                ivUseGridCapThresholdEviction=ObjectGridUtil.isTriggerInEvictionTriggers(Constants.EVICTIONTRIGGER_GRID_CAP, evictionTriggers);
                // grid cap eviction will be triggered by NDO when an insert occurs which exceeds the cap.
                // no need to periodically wake up and check.
                if(ivUseGridCapThresholdEviction) setSleepTime(Integer.MAX_VALUE);
                
            }

            if (ivMaxSize > 0 || ivUseMemoryUsageThresholdEviction || ivUseGridCapThresholdEviction) {
                ivQueues = new Queue[ivNumberOfQueues];
                for (int i = 0; i < ivNumberOfQueues; ++i) {
                    ivQueues[i] = new Queue();
                }
            } else {
                ivNumberOfQueues = 0;
                ivQueues = new Queue[0];
            }
        }

        if (ivUseMemoryUsageThresholdEviction) {
            memoryManagement = MemoryManagementImpl.instance();
            ;
        }

        if (ivUseMemoryEvictionController) {
            this.memEvictionController = MemoryEvictionControllerImpl.instance();
        }
        
    }

    /**
     * Maintains a bi-directional queue ordered on last access.
     *
     * @see Evictor#apply(LogSequence)
     */
    public void apply(LogSequence sequence) {
        final boolean entryEnabled = ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled();
        if (entryEnabled) {
            Tr.entry(tc, "apply");
        }

        KeyType keyMode = ((BaseMap) ivBackingMap).getKeyType();
        // Process LRU data only if size is limited.
        if (ivMaxSize > 0 || ivUseMemoryUsageThresholdEviction || ivUseGridCapThresholdEviction) {
            // we have a maximum size, so see if we need to
            // have evictor thread do something.
            Iterator iter = sequence.getAllChanges();

            // Apply method is called after transaction has completed and locks
            // are released. Thus it is possible that the order the apply methods are called does
            // not match the order in which the transaction completes. For example, an INSERT
            // tran followed by a UPDATE tran might call this method in the reverse order
            // (UPDATE followed by INSERT). Thus we can not depend on LogElement type by itself to
            // determine how to process the LogElement. Instead, we must examine whether the
            // key exists in the BackingMap and if it does, whether the BackingMap entry contains
            // the EvictorData.
            for (int i = sequence.size() - 1; i >= 0; --i) {
                try {
                    LogElement elem = (LogElement) iter.next();
                    CacheEntry entry = ((DiffMapValue)elem).getRawCacheEntry();
                    Object key = elem.getKey();
                    LogElement.Type type = elem.getType();
                    if (type == LogElement.INSERT) {
                        EvictorData evictorData = ivCallback.getEvictorData(key);
                        if (evictorData == null) {
                            // Key does exist in backing map, but CacheEntry for the key
                            // does not contain any EvictorData object. So create and
                            // insert EvictorData into LRU queue and into the CacheEntry
                            // for the key in BackingMap.
                            insertLRU(keyMode, elem, key, entry);

                        } else if (evictorData != EvictorData.KEY_NOT_FOUND) {
                            // Key does exist in BackingMap and the CacheEntry for the key contains
                            // the EvictorData. So update the LRU data.
                            updateLRU(keyMode, elem, key, entry, evictorData);
                        }

                    } else if (type == LogElement.UPDATE || type == LogElement.FETCH || type == LogElement.TOUCH) {
                        EvictorData evictorData = ivCallback.getEvictorData(key);
                        if (evictorData == null) {
                            // Key does exist in backing map and the CacheEntry
                            // for the key
                            // does not contain any EvictorData object. So
                            // create and
                            // insert EvictorData into LRU queue and into the
                            // CacheEntry
                            // for the key in BackingMap.
                            insertLRU(keyMode, elem, key, entry);
                        } else if (evictorData != EvictorData.KEY_NOT_FOUND) {
                            // Key does exist in BackingMap and the CacheEntry
                            // for the key contains
                            // the EvictorData. So update the LRU data.
                            updateLRU(keyMode, elem, key, entry, evictorData);
                        } else {
                            // Key does not exist in BackingMap, so do nothing.
                            // This can happen if
                            // tran 1 does a INSERT and tran 2 and UPDATE, but
                            // apply for each tran
                            // is called in reverse order. If that happens, then
                            // when INSERT tran
                            // calls apply we will insert the LRU data into
                            // queue at that time.
                        }
                    } else if (type == LogElement.DELETE || type == LogElement.EVICT) {
                        // Ensure LRU queue no longer contains evictor data for this LogElement.
                        // getEvictorData will do what is below.  Being an internal plug-in we can do it faster.
                        DiffMapValue dmv = (DiffMapValue) elem;
                        SystemCacheEntry beforeImage = (SystemCacheEntry) dmv.getRawBeforeImage();
                        EvictorData evictorData = (beforeImage == null) ? null : beforeImage.getEvictorData();
                        if (evictorData != null && evictorData != EvictorData.KEY_NOT_FOUND) {
                            removeLRU(keyMode, elem, entry, evictorData);
                        }
                    } else if (type == LogElement.CLEAR) {
                        if (disk || offheap) {
                            delegator.truncateTable();
                            for (int idx = 0; idx < ivNumberOfQueues; ++idx) {
                                synchronized (queueCounts[idx]) {
                                    queueCounts[idx].count = 0;
                                }
                            }
                        } else if ((ivQueues != null) && (ivNumberOfQueues > 0)) {
                            for (int idx = 0; idx < ivNumberOfQueues; ++idx) {
                                synchronized (ivQueues[idx]) {
                                    ivQueues[idx].clear();
                                }
                            }
                        }
                    } else if (type == LogElement.LOCK) {
                        //  No action. Handled by locking code only.
                    } else {
                        // unknown type, dump information, and continue...
                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                            Tr.debug(tc, "Unknown type of LogElement: " + type);
                        }

                        if (ivLogInvalidLogElement) {
                            // Only log the first time this happens.
                            ivLogInvalidLogElement = false;

                            // CWOBJ0012E: The LogElement type code, {0} ({1}),
                            // is not recognized for this operation.
                            Integer code = new Integer(type.getCode());
                            String string = type.toString();
                            Tr.error(tc, NLSConstants.INVALID_LOGELEMENT_TYPE_CWOBJ0012, new Object[] { code, string });
                        }
                    }
                    
                } catch (Throwable t) {
                    // We do not expect this to ever happen, but if it does, we
                    // simply FFDC and
                    // log warning message. We will continue with the loop so
                    // that we have a
                    // chance to process the rest of the LogElements in the
                    // LogSequence.
                    FFDCFilter.processException(t, CLASS_NAME + ".apply", "314", this);
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, "apply caught unexpected Throwable", t);
                    }
                    Tr.warning(tc, NLSConstants.IGNORING_UNEXPECTED_EXCEPTION_CWOBJ0002,
                            new Object[] { t, "LRUEvictor" });
                }
            }
        }

        if (entryEnabled) {
            Tr.exit(tc, "apply");
        }
    }

    /**
     * Insert LRU queue element into the LRU queue for a specified key and CacheEntry. This method MUST be called only
     * when key exists in BackingMap and CacheEntry does NOT contain any EvictorData object.
     *
     * @param key
     *            is the key of BackingMap entry.
     * @param cacheEntry
     *            is the CacheEntry that currently does NOT contain any EvictorData object.
     */
    private void insertLRU(KeyType keyMode, LogElement elem, Object key, CacheEntry cacheEntry) throws Exception {
        final boolean entryEnabled = ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled();
        if (entryEnabled) {
            Tr.entry(tc, "insertLRU for key: " + key);
        }

        // Create evictor data for CacheEntry in BackingMap.
        LRUEvictorData data;
        if (disk) {
            data = new DiskLRUEvictorData(key, elem.getLastAccessTime());
        } else if (offheap) {
            ObjectGridHashTableOH oht = (ObjectGridHashTableOH) ((BaseMap) ivBackingMap).getCacheEntries();
            XsOffHeapEvictionDataImpl xsEvictData = oht.createXsOffHeapEvictionData(key, cacheEntry); //offheap: allocation pin
            data = new LRUOffHeapEvictionData((SystemMap)ivBackingMap, xsEvictData); 
        } else
            data = new LRUEvictorData(key);
        boolean entryInserted = true;
        try {
            ivCallback.setEvictorData(key, data); //offheap: pinned by XsValue
        } catch (IllegalArgumentException iae) {
            entryInserted = false;
        }

        if ((disk || offheap) && entryInserted) {
            int queue = determineEntryQueue(cacheEntry);
            synchronized (queueCounts[queue]) {
                delegator.insertData(data, key, cacheEntry, queue);
                queueCounts[queue].count++;
            }
        } else if (entryInserted) {
            // Get Queue to use for this key.
            Queue queue = getQueue(cacheEntry);

            // Add LRUQueueEntry to the queue.
            synchronized (queue) {
                // Small timing window between threads. Need to make sure
                // another
                // thread didn't add it first.

                // putFirst will move it to be the first or
                // will add it to be first
                queue.putFirst(data);
            }
        }
        
        
        // Offheap: We have completed the insertion of the evictData into the queue, with
        //  a new of four pins (allocation, xsValue->evictData, queue->first/next, queue->last/prev).
        //  At this point, we release the allocation pin.
        if (offheap) 
        	OffHeapManager.getInstance().unpinXsEvictionData(
        			((LRUOffHeapEvictionData) data).getXsOffHeapEvictionData().getAddress(), 
        			OffHeapManager.JAVA_GET_PIN);

        if (entryEnabled) {
            Tr.exit(tc, "insertLRU for key: " + key);
        }
    }

    /**
     * This method is used to update the LRU queue when both the key exists in BackingMap AND the CacheEntry does
     * contain the EvictorData object for the key.
     *
     * @param key
     * @param cacheEntry
     * @param evictorData
     */
    private void updateLRU(KeyType keyMode, LogElement elem, Object key, CacheEntry cacheEntry, EvictorData evictorData) throws Exception {
        final boolean entryEnabled = ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled();
        if (entryEnabled) {
            Tr.entry(tc, "updateLRU for key: " + key);
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "updateLRU moving cache entry to head of LRU queue for key = " + key);
        }

        if (disk) {
            int queue = determineEntryQueue(cacheEntry);

            DiskLRUEvictorData data = (DiskLRUEvictorData) evictorData;
            data.setLastAccessTime(elem.getLastAccessTime());
            synchronized (queueCounts[queue]) {
                delegator.updateData(data, key, cacheEntry, queue);
            }
        } else if (offheap) {
            int queue = determineEntryQueue(cacheEntry);

            LRUOffHeapEvictionData data = (LRUOffHeapEvictionData) evictorData;
            synchronized (queueCounts[queue]) {
                delegator.updateData(data, key, cacheEntry, queue);
            }
        } else {
            // Key found and LRUQueueEntry found. So move LRUQueueEntry to head of
            // LRU queue.
            LRUEvictorData data = (LRUEvictorData) evictorData;

            // Get Queue to use for this key.
            Queue queue = getQueue(cacheEntry);
            synchronized (queue) {
                // Small timing window between evictor and this thread. Need
                // to make sure it is on the queue before removing it.

                // putFirst will move it to be the first or
                // will add it to be first
                queue.putFirst(data);
            }

            if (entryEnabled) {
                Tr.exit(tc, "updateLRU for key: " + key);
            }
        }
    }

    /**
     * @see RollbackEvictor#rollingBack(LogSequence)
     */
    public void rollingBack(LogSequence sequence) {
        final boolean entryEnabled = ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled();
        if (entryEnabled) {
            Tr.entry(tc, "rollingBack");
        }

        KeyType keyMode = ((BaseMap) ivBackingMap).getKeyType();
        Iterator iter = sequence.getAllChanges();

        for (int i = sequence.size() - 1; i >= 0; --i) {
            try {
                // Get next LogElement.
                LogElement elem = (LogElement) iter.next();

                if (elem.getType() != LogElement.CLEAR) {
                    // Get the evictor data from the CacheEntry.
                    CacheEntry cacheEntry = ((DiffMapValue)elem).getRawCacheEntry();
                    Object key = elem.getKey();
//                    if (keyMode.isBytes()) {
//                        // TODO DS: API gets Key wrapper or byte[]? Delete this if Key wrapper is the default?
//                        key = ((KeyWrapper)cacheEntry).getKeyRaw();
//                    }
                    EvictorData evictorData = ivCallback.getEvictorData(key);
                    if (evictorData != null && evictorData != EvictorData.KEY_NOT_FOUND) {
                        // key does exist in BackingMap and CacheEntry contains
                        // EvictorData,
                        // so update the LRU queue to reflect access to this
                        // key.
                        updateLRU(keyMode, elem, key, cacheEntry, evictorData);
                    }
                }
            } catch (Throwable t) {
                FFDCFilter.processException(t, CLASS_NAME + ".rollback", "594", this);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                    Tr.debug(tc, "rollback caught unexpected Throwable", t);
                }
                Tr.warning(tc, NLSConstants.IGNORING_UNEXPECTED_EXCEPTION_CWOBJ0002, new Object[] { t, "LRUEvictor" });
            }
        }

        if (entryEnabled) {
            Tr.exit(tc, "rollingBack");
        }
    }

    /**
     * Remove the LRUQueueEntry object from the LRU queue that is associated with a specified LogElement of type DELETE
     * or EVICT.
     *
     * @param elem
     *            is a LogElement where elem.getType is LogElement.DELETE or LogElement.EVICT.
     */
    private void removeLRU(KeyType keyMode, LogElement elem, CacheEntry cacheEntry, EvictorData evictorData) throws Exception {
        // Use the before image from LogElement to find the EvictorData
        // object that needs to be removed from the LRU queue. If evictor
        // thread evicted entry, then the before image will be null and there
        // is no need to remove from LRU queue since it was already removed
        // by the evictor thread.
        // Object key = elem.getCacheEntry().getKey();

        final boolean entryEnabled = ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled();
        if (entryEnabled) {
            Tr.entry(tc, "removeLRU for LRUEvictorData: " + evictorData);
        }
        // Don't make an additional trip to check for presence of evData...just
        // do another remove
        if (disk || offheap) {
            int queue = determineEntryQueue(cacheEntry);
            QueueCount cur = queueCounts[queue];
            synchronized(cur){
                delegator.removeData(evictorData, queue);
                if (cur.count > 0) cur.count--;
            }
        } else {
            // Get Queue to use for the key that is in CacheEntry.
            Queue queue = getQueue(cacheEntry);
            LRUEvictorData lruData = (LRUEvictorData) evictorData;
            synchronized (queue) {
                // If evictor thread, it is already removed, so we need
                // to make sure it is on queue before removing.
                if (lruData.onQueue()) {
                    queue.remove(lruData);
                }
            }
        }

        if (entryEnabled) {
            Tr.exit(tc, "removeLRU");
        }
    }

    /**
     * @param entry
     *            the cahce entry
     * @return Queue to use for a specified CacheEntry.
     */
    private Queue getQueue(CacheEntry entry) {
        // Hash the CacheEntry hashCode to determine which BinaryHeap to use
        // for this CacheEntry. This should give us a good distribution
        // across the BinaryHeap.
        int queueId = determineEntryQueue(entry);
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "getQueue using queue in bucket " + queueId);
        }
        Queue queue = ivQueues[queueId];
        return queue;
    }

    private int determineEntryQueue(CacheEntry entry) {
        int hashCode = entry.hashCode();
        return XSUtilities.computeHash(hashCode, divisor) % ivNumberOfQueues;
    }

    /**
     * Gets the maximum size of each LRU queue.
     *
     * @return the same value that was passed to the <code>setMaxSize(int)</code> method or the default value of zero if
     *         the <code>setMaxSize</code> method is never called
     *
     * @see #setMaxSize(int)
     */
    public int getMaxSize() {
        return ivMaxSize;
    }

    /**
     * Sets the maximum size of each LRU queue.
     * <p>
     * This method is used to override the default maximum size for each LRU queue used to keep LRU data. The evictor
     * thread will attempt to keep each LRU queue to be no larger than the maximum size. This method must be called
     * prior to the <code>initialize</code> method to avoid an IllegalStateException being thrown.
     *
     * @param maxSize
     *            is the maximum size per LRU queue. Any value <= 0 indicates to allow each queue to be of unlimited
     *            size. In which case, no LRU data is kept.
     *
     * @throws IllegalStateException
     *             if called after the initialize method.
     *
     * @see #initialize(BackingMap, EvictionEventCallback)
     */
    public void setMaxSize(int maxSize) {
        if (ivQueues != null) {
            throw new IllegalStateException("must be called prior to initialize method.");
        }

        ivMaxSize = maxSize;
    }

    /**
     * Gets the sleep time being used in seconds.
     *
     * @return the same value that was passed to the <code>setSleepTime(int)</code> method or the default value of
     *         <code>DEFAULT_SLEEP_TIME</code> if the <code>setSleepTime</code> method is never called
     *
     * @see #setSleepTime(int)
     */
    public int getSleepTime() {
        return (int) (ivSleepTime / 1000L);
    }

    /**
     * Sets the sleep time to use in seconds.
     * <p>
     * This method is used to override the default sleep time of the evictor thread in seconds. This method must be
     * called prior to the <code>initialize</code> method to avoid an IllegalStateException being thrown. If this method
     * is not called, the <code>DEFAULT_SLEEP_TIME</code> constant is used as the sleep time.
     *
     * @param seconds
     *            is the number of seconds the evictor thread sleeps in between each sweep of the LRU queue data
     *
     * @throws IllegalArgumentException
     *             if seconds < 1
     * @throws IllegalStateException
     *             if called after the initialize method.
     *
     * @see #DEFAULT_SLEEP_TIME
     * @see #initialize(BackingMap, EvictionEventCallback)
     */
    public void setSleepTime(int seconds) {
        if (seconds < 1) {
            throw new IllegalArgumentException("seconds argument must be > 0");
        }

        if (ivQueues != null) {
            throw new IllegalStateException("must be called prior to initialize method.");
        }

        ivSleepTime = seconds * 1000L;
    }

    /**
     * Gets number of LRU queues being used.
     *
     * @return the same value that was passed to the <code>setNumberOfLRUQueues(int)</code> method or the default value
     *         of <code>DEFAULT_NUMBER_OF_QUEUES</code> if the <code>setNumberOfLRUQueues</code> method is never called
     *
     * @see #DEFAULT_NUMBER_OF_QUEUES
     * @see #setNumberOfLRUQueues(int)
     */
    public int getNumberOfLRUQueues() {
        return ivNumberOfQueues;
    }

    /**
     * Sets the number of LRU queues to use.
     * <p>
     * This method is used to override the default number of LRU queues created by the <code>initialize</code> method
     * for keeping the LRU data needed by the evictor thread. This method must be called prior to the
     * <code>initialize</code> method to avoid an IllegalStateException being thrown. If this method is not called,the
     * <code>DEFAULT_NUMBER_OF_QUEUES</code> constant is used as the number of heaps.
     *
     * @param numberOfQueues
     *            is the number of LRU queue instances used to hold LRU data. The value must be greater than or equal to
     *            1 and a prime number is recommended for best performance.
     *
     * @throws IllegalArgumentException
     *             if numberOfQueuess < 1
     * @throws IllegalStateException
     *             if called after the initialize method.
     *
     * @see #DEFAULT_NUMBER_OF_QUEUES
     * @see #initialize(BackingMap, EvictionEventCallback)
     */
    public void setNumberOfLRUQueues(int numberOfQueues) {
        if (numberOfQueues < 1) {
            throw new IllegalArgumentException("numberOfQueues argument must be > 0");
        }

        if (ivQueues != null) {
            throw new IllegalStateException("must be called prior to initialize method.");
        }

        ivNumberOfQueues = numberOfQueues;
    }


    private void doMemoryEviction() {
        double recommendedPercentToEvict = memEvictionController.getEvictionRecommendation();
        evictPercent(recommendedPercentToEvict);
    }

    private void doGridCapEviction() {
    	if(tc.isEntryEnabled()) {
    		Tr.entry(tc,"doGridCapEviction");
    	}    	
    	boolean isOverflow=((ServerPropertiesImpl)ServerFactory.getServerProperties()).isDiskOverflow() || ((ServerPropertiesImpl)ServerFactory.getServerProperties()).isDiskOverflowEnabled();
    	if(!isOverflow) return;
    	double recommendedPercentToEvict = DiskHelperFactory.getDiskHelper().getEvictionRecommendation(mapSetId);
        if(recommendedPercentToEvict>0) {
        	// if we're going to bother evicting because we exceed the recommended size,
        	// evict an extra 5% while we're at it.
        	recommendedPercentToEvict+=0.05;
        	if(tc.isDebugEnabled()) Tr.debug(tc,"Evicting "+recommendedPercentToEvict*100+"%");
            evictPercent(recommendedPercentToEvict);
        }
    	if(tc.isEntryEnabled()) {
    		Tr.exit(tc,"doGridCapEviction");
    	}

    }

    private void evictPercent(double percent) {
        boolean traceSize = ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled();
        long totalSize = 0;
        if(offheap) {
        	try {
	        	LRUOffHeapEvictionDelegator lfe = (LRUOffHeapEvictionDelegator) delegator;
	            int[] counts=lfe.getQueueCounts();               
	            int currentSize=0;
	            for(int i=0;i<counts.length;i++) {
	         	   currentSize+=counts[i];
	            }
	            int targetSize=(int)(currentSize-percent*currentSize);
	            if (ivMaxSize > 0 && ivMaxSize < targetSize) {
	                 targetSize = ivMaxSize;
	            }
	            int totalEntriesToEvict=currentSize-targetSize;
	            if(totalEntriesToEvict<=0) return;
	            int entriesPerQueue=(int)Math.ceil(totalEntriesToEvict/(double)ivNumberOfQueues);
	            if(traceSize)
	            	Tr.debug(tc, "currentSize="+currentSize+" percent=" + percent +" targetSize="+targetSize+" entriesPerQueue="+entriesPerQueue);
	            List evictList = null;
	            for (int i = 0; i < ivNumberOfQueues; ++i) {
	            	evictList = delegator.getEvictList(i, entriesPerQueue); //Offheap: pins each entry in List
	                ivCallback.evictMapEntries(evictList);
	                
	                // Offheap: Release the pins from the getEvictList call	
	                if(!evictList.isEmpty()) {
	                 	for(OffHeapEvictionData ohed : (List<OffHeapEvictionData>)evictList) {
	                		OffHeapManager.getInstance().unpinXsEvictionData(ohed.getAddress(), OffHeapManager.JAVA_GET_PIN);
	                	}
	                } 
	            }
	
	    	}catch(Exception e) {
                FFDCFilter.processException(e, CLASS_NAME + ".evictPercent", "909", evictorId);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                    Tr.debug(tc, evictorId + " caught unexpected Throwable: " + e, e);
                }

                // CWOBJ0002W: ObjectGrid component is ignoring
                // unexpected exception: {0}.
                Tr.warning(tc, NLSConstants.IGNORING_UNEXPECTED_EXCEPTION_CWOBJ0002, new Object[] { e,"LRUEvictor" });

        	}
        } else {
	        for (int i = 0; i < ivNumberOfQueues; ++i) {
	            try {
	                Queue queue = ivQueues[i];
	                List evictList = null;
	                synchronized (queue) {
	                    long size = queue.size();
	                    long targetSize = (long) (size - percent * size);
	                    if (ivMaxSize > 0 && ivMaxSize < targetSize) {
	                        targetSize = ivMaxSize;
	                    }
	
	                    if (size > targetSize) {
	                        evictList = new ArrayList((int) (size - targetSize));
	                        while (queue.size() > targetSize) {
	                            LRUEvictorData data = (LRUEvictorData) queue.removeLast();
	                            if (data != null) {
	                                evictList.add(data);
	                            }
	                        }
	                    }
	                    totalSize += queue.size();
	                }
	
	                if (evictList != null && evictList.size() > 0) {
	                    ivCallback.evictMapEntries(evictList);
	
	                }
	            } catch (Throwable t) {
	                // Should never happen, but let's make sure evictor does not die
	                // unexpectantly by tolerating all Throwable objects. Just log
	                // the fact that it occured.
	                FFDCFilter.processException(t, CLASS_NAME + ".evictPercent", "924");
	
	                // CWOBJ0002W: ObjectGrid component is ignoring unexpected
	                // exception: {0}.
	                Tr.warning(tc, NLSConstants.IGNORING_UNEXPECTED_EXCEPTION_CWOBJ0002, new Object[] { t, "LRUEvictor" });
	            }
	        }
        }
        if (traceSize) {
            Tr.debug(tc, evictorId + " total number of LRUEvictorData objects in LRU queues = " + totalSize);
        }
    	
    }
    
    /**
     * Manages the map's size and sends eviction events to the Map.
     * <p>
     * The thread is only spawned if the map size is limited rather than unlimited in size (e.g. maxSize > 0 ).
     */
    public void run() {
        synchronized (this) {
//        	try{Thread.sleep(6000000);} catch(Exception e){}
            scheduledEviction = null;
            if (activeThread == null) {
                activeThread = Thread.currentThread();
            } else {
                return;
            }
        }

        Thread currentThread = Thread.currentThread();
        String currentThreadName = currentThread.getName();
        currentThread.setContextClassLoader(classLoader);
        currentThread.setName(currentThreadName + ": " + evictorId);

        try {
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
                Tr.entry(tc, "run", evictorId + " running, countHandleManagementEvent = "
                        + (listener == null ? 0 : listener.getMemoryEvictionCount()));
            }

            // Loop until destroy or deactivate is called or we break out of the
            // loop.
            while (!ivQuit) {

                // new code path for memory based eviction
                if (this.ivUseMemoryEvictionController) {
                    doMemoryEviction();
                } else if(this.ivUseGridCapThresholdEviction) {
                	doGridCapEviction();
                    // original code path is intact in the else section
                } else {
                    // when ivUseMemoryUsageThresholdEviction is true,
                    // after wake up, need to check if it is notified by
                    // MemoryNotification event.
                    // if YES, need to evict entries even the current size has
                    // not exceeded max size.
                    boolean isUsageThresholdExceeded = false;
                    boolean offheapAndMemoryUsageThreshold=false;
                    int entriesPerQueue=0;
                    if (ivUseMemoryUsageThresholdEviction ) {
                        if (!offheap){
                            isUsageThresholdExceeded = memoryManagement.isUsageThresholdExceeded();
                   
                        }else{
                            offheapAndMemoryUsageThreshold=true;
                            
                            LRUOffHeapEvictionDelegator lfe = (LRUOffHeapEvictionDelegator)delegator;
                            int totalEntriesToEvict = lfe.totalEntriesToEvict();
                            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                                Tr.debug(tc, "run", evictorId + " offheapAndMemoryUsageThreshold="+offheapAndMemoryUsageThreshold+" totalEntriesToEvict="+totalEntriesToEvict);
                            }
                            if (totalEntriesToEvict>0){
                                isUsageThresholdExceeded=true;
                                entriesPerQueue=(int)Math.ceil(totalEntriesToEvict/ivNumberOfQueues);
                            }
                            
                        }
                    }
                    // Process data in array of LRU Queue objects.
                    int totalSize = 0;
                    boolean traceSize = ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled();
                    for (int i = 0; i < ivNumberOfQueues; ++i) {
                        try {
                            if (disk || offheap) {
                                if (offheap && ((ivUseMemoryUsageThresholdEviction && !isUsageThresholdExceeded))) {
                                    break;
                                }

                                if (offheapAndMemoryUsageThreshold) {
                                    //this means it is offheap and the memory threshold has been configured
                                    List evictList = null;
                                    if (isUsageThresholdExceeded) {
                                        //threshold has been breached
                                        LRUOffHeapEvictionDelegator lfe = (LRUOffHeapEvictionDelegator) delegator;
                                       
                                        evictList = delegator.getEvictList(i, entriesPerQueue); //Offheap: pins each entry in List
                                    } else {
                                        //go back to sleep if threshold not exceeded
                                        break;
                                    }

                                    ivCallback.evictMapEntries(evictList);
                                    
                                    // Offheap: Release the pins from the getEvictList call
                                    if(!evictList.isEmpty()) {
                                    	for(OffHeapEvictionData ohed : (List<OffHeapEvictionData>)evictList) {
                                    		OffHeapManager.getInstance().unpinXsEvictionData(ohed.getAddress(), OffHeapManager.JAVA_GET_PIN);
                                    	}
                                    }
                                    
                                    continue;
                                }

                                //get tableSize from memory instead of delegator.getTableSize();
                                List evictList = null;
                                synchronized (queueCounts[i]) {
                                    long tableSize = queueCounts[i].count;
                                    // Figure out num to evict
                                    long numToEvict = 0;
                                    if (ivMaxSize > 0) {
                                        numToEvict = (tableSize > ivMaxSize) ? tableSize - ivMaxSize : 0;
                                    }

                                    if (numToEvict > 0) {
                                        evictList = delegator.getEvictList(i, numToEvict); //Offheap: pins each entry in List
                                    }
                                }

                                if (evictList != null && !evictList.isEmpty()) {
                                    if (tc.isDebugEnabled()) {
                                        Tr.debug(tc, "LRUEvictor evicting: " + evictList);
                                    }
                                    ivCallback.evictMapEntries(evictList);
                                    
                                    // Offheap: Release the pins from the getEvictList call
                                    if(offheap) {
                                    	for(OffHeapEvictionData ohed : (List<OffHeapEvictionData>)evictList) {
                                    		OffHeapManager.getInstance().unpinXsEvictionData(ohed.getAddress(), OffHeapManager.JAVA_GET_PIN);
                                    	}
                                    }
                                    
                                }
                            } else {
                                Queue queue = ivQueues[i];
                                processQueue(queue, isUsageThresholdExceeded);
                                synchronized (queue) {
                                    totalSize += queue.size();
                                }
                            }
                        } catch (Throwable t) {
                            // Should never happen, but let's make sure evictor
                            // does not die
                            // unexpectantly by tolerating all Throwable
                            // objects. Just log
                            // the fact that it occured.
                            FFDCFilter.processException(t, CLASS_NAME + ".run", "365", evictorId);
                            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                                Tr.debug(tc, evictorId + " caught unexpected Throwable: " + t, t);
                            }

                            // CWOBJ0002W: ObjectGrid component is ignoring
                            // unexpected exception: {0}.
                            Tr.warning(tc, NLSConstants.IGNORING_UNEXPECTED_EXCEPTION_CWOBJ0002, new Object[] { t,
                                    "LRUEvictor" });
                        }
                    }

                    if (traceSize) {
                        Tr.debug(tc, evictorId + " total number of LRUEvictorData objects in LRU queues = " + totalSize);
                    }
                }

                // resume original code path to put the thread to sleep
                try {
                    // Sleep for a while before sweeping over BinaryHeap list.
                    // Thread.sleep( ivSleepTime );

                    boolean goToSleep = true;
                    if (ivUseMemoryUsageThresholdEviction) {
                        // Need to yield a bit here, or we may never give the VM
                        // a chance
                        // to update the usage count, which means we'd evict
                        // everything.
                        Thread.sleep(100);

                        // before go to waiting state, ensure
                        // isUsageThresholdExceeded is false when
                        // ivUseMemoryUsageThresholdEviction
                        // if the map size is 0 or we're no longer exceeding the usage threshold, finish and reschedule for later (goToSleep)
                        goToSleep = (((BaseMap)ivBackingMap).size() == 0 || !memoryManagement.isUsageThresholdExceeded());
                    }

                    if (goToSleep) {
                        // break out of the while loop and schedule the next
                        // eviction
                        break;
                    }
                } catch (InterruptedException e) {
                    // FFDCFilter.processException(t, CLASS_NAME + ".run",
                    // "353", threadName);
                    // ignore exception.
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, evictorId + " was interrupted");
                    }
                }
            }

            synchronized (this) {
                if (activeThread == Thread.currentThread()) {
                    activeThread = null;
                }
                if (!ivQuit) {
                    scheduledEviction = threadPool.schedule(this, ivSleepTime);
                } else {
                    if (ivDestroyed) {
                        if (disk || offheap) {
                            delegator.dropTable();
                            queueCounts = null;
                        }
                        ivQueues = null;
                        ivNumberOfQueues = 0;
                    }
                }
                notifyAll();
            }

            if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
                Tr.exit(tc, "run", evictorId + " terminating, countHandleManagementEvent = "
                        + (listener == null ? 0 : listener.getMemoryEvictionCount()));
            }
        } finally {
            currentThread.setName(currentThreadName);
        }

    } // end run

    // this method is never called in this class, but is called via reflection
    // from MemoryEvictionTrigger
    // this method must remain private it is a external class
    private synchronized void triggerEviction() {
        if (!ivQuit && activeThread == null) {
            // Cancel the scheduledEviction if there is one
            if (scheduledEviction != null) {
                scheduledEviction.cancel();
            }
            threadPool.execute(this);
        }
    }

    /**
     * The evictor must manage it's size here and send eviction events to the Map. The thread is only spawned if the map
     * size is limited rather than unlimited in size (e.g. maxSize > 0 ).
     *
     * @param queue
     */
    private void processQueue(Queue queue, boolean isUsageThresholdExceeded) {
        // This thread continues to run until destroy method is called.
        int queueSize = 0;
        synchronized (queue) {
            queueSize = queue.size();
            if (!isUsageThresholdExceeded && ivMaxSize == 0) {
                // not in critical memory usage and ivMaxSize == 0 (unlimited
                // size)
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                    Tr.debug(tc, evictorId + " processQueue: " + queue + ", isUsageThresholdExceeded: "
                            + isUsageThresholdExceeded + ", ivMaxSize: " + ivMaxSize + ", No need to evict");
                }
                return;
            }

            // cheng1: should just return if queueSize == 0 ?
            if (queueSize <= 0) {
                return;
            }
        }

        // Anything need to be evicted?
        // int numToEvict = queueSize - ivMaxSize;
        // int numToEvict = (queueSize > ivMaxSize) ? queueSize - ivMaxSize : 0;

        int numToEvict = 0;
        if (ivMaxSize > 0) {
            // normal mode
            numToEvict = (queueSize > ivMaxSize) ? queueSize - ivMaxSize : 0;
        }
        // else, the numToEvict = 0; when UsageThresholdExceeded mode or
        // ivMaxSize==0

        if (numToEvict == 0 && isUsageThresholdExceeded) {
            // even the heap size has not reached max size, but the memory
            // exceeded event has occurred,
            // still need to evict entries to avoid OutOfMemory error.
            // evict 3% for now.
            // to do: find a better logic to determine the number to evict for
            // various situations
            double temp = queueSize * 0.03;
            numToEvict = (int) temp;

            if (numToEvict == 0) {
                // there is not many entries, less than 34.
                if (queueSize >= 20) {
                    numToEvict = 3;
                } else if (queueSize >= 10) {
                    numToEvict = 2;
                } else {
                    numToEvict = 1;
                }
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, evictorId + " processQueue: " + queue + ", isUsageThresholdExceeded: "
                    + isUsageThresholdExceeded + ", ivMaxSize: " + ivMaxSize + ", numberToEvict: " + numToEvict);
        }

        if (numToEvict > 0) {
            // Yep, start at tail of queue and evict the first
            // numToEvict entries since they are the least
            // recently used entries.
            List evictList = new ArrayList(numToEvict);
            if (queueSize > ivMaxSize) {
                while (queueSize > ivMaxSize) {
                    LRUEvictorData data = null;
                    synchronized (queue) {
                        // avoid the NoSuchElementException by checking if the queue is empty when we have the sync block.
                        if (!queue.isEmpty()) {
                            data = (LRUEvictorData) queue.removeLast();
                        }
                        queueSize = queue.size();
                    }

                    if (data != null) {
                        evictList.add(data);
                    }
                }
            } else if (numToEvict > 0) {
                // check queueSize again, in case it has been modified by other
                // threads
                queueSize = queue.size();
                while (queueSize > 0 && numToEvict > 0) {
                    LRUEvictorData data = null;
                    synchronized (queue) {
                        // avoid the NoSuchElementException by checking if the queue is empty when we have the sync block.
                        if (!queue.isEmpty()) {
                            data = (LRUEvictorData) queue.removeLast();
                        }
                        queueSize = queue.size();
                    }

                    if (data != null) {
                        evictList.add(data);
                        --numToEvict;
                    }
                }

            }

            // Callback to actually evict the entries if there is anything
            // added to the keysToEvict list.
            try {
                if (!evictList.isEmpty()) {
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, "LRUEvictor evicting: " + evictList);
                    }
                    ivCallback.evictMapEntries(evictList);
                }
            } catch (Throwable ex) {
                // FFDCFilter.processException(ex, CLASS_NAME +
                // ".processEvictorHeap", "635", this );

                // No Throwable are expected, but we are going to tolerate any
                // that does
                // occur to ensure that this loop continues to process every key
                // in the
                // eviction list.
            }
        } // end if ( numToRemove > 0 )

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "processQueue");
        }
    }

    /**
     * @see Evictor#destroy()
     */
    synchronized public void destroy() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "destroy: " + evictorId);
        }
        ivDestroyed = true;
        deactivate(ObjectGridImpl.DESTROY_DEFAULT);
    }

    synchronized private void destroy(int destroyType) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "destroy: " + ObjectGridImpl.getDestroyType(destroyType) + " " + evictorId);
        }
        ivDestroyed = true;
        deactivate(destroyType);
    }

    /**
     * @see Evictor#activate()
     */
    synchronized public void activate() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "activate: " + evictorId + ", MemoryUsageThresholdEviction: "
                    + ivUseMemoryUsageThresholdEviction);
        }

        if (!ivActivated) {
            ivActivated = true;
            if (ivQuit && (ivMaxSize > 0 || ivUseMemoryUsageThresholdEviction || ivUseGridCapThresholdEviction)) {
                ivQuit = false;
                threadPool.execute(this);
            }

            if (!disk && !offheap) {
                if (!ivQuit && ivUseMemoryUsageThresholdEviction) {
                    if (!memoryManagement.isUsageThresholdSupported()) {
                        if (!hasOutputUsageThresholdNotSupportedMsg) {
                            // there might be thousands evictors activate and
                            // deactivate
                            // only out put warning once, it will show something
                            // like:
                            // LRUEvictor W CWOBJ0035W: Memory utilization threshold
                            // not supported for this JVM.
                            Tr.warning(tc, NLSConstants.MEMORY_UTILIZATION_THRESHOLD_NOT_SUPPORTED_CWOBJ0035);
                            hasOutputUsageThresholdNotSupportedMsg = true;
                        }
                        ivUseMemoryUsageThresholdEviction = false;
                    } else {
                        // register listener with MemoryManagement

                        listener = new MemoryEvictionTrigger(triggerEvictionMethod, this, "LRU");
                        memoryManagement.addEventHandler(listener);
                    }
                }

                // new memory eviction. This cannot be enabled if the old mechanism
                // is enabled
                if (ivUseMemoryEvictionController && !ivQuit) {
                    listener = new MemoryEvictionTrigger(triggerEvictionMethod, this, "LRU");
                    memEvictionController.addEventHandler(this, this.ivBackingMap, listener);
                }
            }
            
            if(ivUseGridCapThresholdEviction) {
            	String mapName=ivBackingMap.getName();
            	String gridName=ivBackingMap.getObjectGrid().getName();
            	int partition=ivBackingMap.getPartitionId();
            	
                DiskHelper diskHelper = DiskHelperFactory.getDiskHelper();
            	mapId = DiskOverflowMapIdentifier.getMapId(gridName, mapName, partition);
            	mapSetId = DiskOverflowMapIdentifier.getMapSetId(gridName, ivBackingMap.getMapSetName(), partition);
            	diskHelper.registerCapEvictorMethod(mapId, this, triggerEvictionMethod);
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "activate");
        }
    }

    /**
     * @see Evictor#deactivate()
     */
    synchronized public void deactivate() {
        deactivate(ObjectGridImpl.DESTROY_DEFAULT);
    }

    synchronized private void deactivate(int destroyType) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "deactivate: " + ObjectGridImpl.getDestroyType(destroyType) + " " + evictorId);
        }

        if (!ivQuit) {
            ivQuit = true;
            // Cancel the scheduledEviction if there is one
            if (scheduledEviction != null) {
                scheduledEviction.cancel();
            }
        }

        if (listener != null) {
            if (ivUseMemoryUsageThresholdEviction) {
                // remove from MemoryManagement
                memoryManagement.removeEventHandler(this.listener);
            } else if (this.ivUseMemoryEvictionController) {
                memEvictionController.removeEventHandler(this, this.ivBackingMap, this.listener);
            }
        }

        if(ivUseGridCapThresholdEviction) {
            DiskHelperFactory.getDiskHelper().deregisterCapEvictorMethod(mapId);
        }

        
        ivActivated = false;
        if (activeThread != null) {
            try {
                wait();
            } catch (InterruptedException e) {
                // fine
            }
        }

        if (ivDestroyed) {
            // Deactivate called by destroy method, so destroy array
            // of Queues that holds the evictor data if evictor
            // thread is not running. If evictor thread is running,
            // let the evictor thread destroy binary heap objects
            // before the thread terminates.
            if (ivQuit) {
                // if this is a DefaultGrid map, don't drop the table, we don't do that for DefaultGrid so we
                // can hibernate the data.
                if ((disk || offheap) && destroyType != ObjectGridImpl.DESTROY_MEMORY_ONLY
                        && (ivBackingMap == null || (!ivBackingMap.getObjectGrid().getName().equals("DefaultGrid") &&
						 						     !ivBackingMap.getObjectGrid().getName().equals("_IBM_XC10_CONFIG_"))
                            )) {
                    delegator.dropTable();
                    queueCounts = null;
                }
                ivNumberOfQueues = 0;
                ivQueues = null;
            }
        } else {
            // We are deactivating, but not destroying LRUEvictor. We need
            // to clear the Queue of data since the BaseMap was cleared
            // of all cache entries as part of the deactivate.
            if (disk || offheap) {
                delegator.truncateTable();
                if (queueCounts != null) {
                    for (int i = 0; i < ivNumberOfQueues; ++i) {
                        synchronized (queueCounts[i]) {
                            queueCounts[i].count = 0;
                        }
                    }
                }
            } else {
                if ((ivQueues != null) && (ivNumberOfQueues > 0)) {
                    for (int i = 0; i < ivNumberOfQueues; ++i) {
                        ivQueues[i].clear();
                    }
                }
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "deactivate");
        }
    }
}