// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.7 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/security/SecurityConstants.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:24:59
//
// Module  :  SecurityConstants.java
//
// Source File Description:
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   04/12/2005   jitang    ObjectGrid - JAAS / Security - Drop 1 - Basic Support
// d329154      WASX.XD   12/05/2005   jitang    Open up SecureToken interface.
// PK17543      WASX.XD   01/26/2006   jitang    Update JavaDoc
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security;

import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.security.config.ClientSecurityConfiguration;

/**
 * This class contains the constants used for security configuration.
 *
 * @since WAS XD 6.0, XC10
 */
public class SecurityConstants {

    /**
     * Constant representing JAAS authorization
     *
     * @see ObjectGrid#setAuthorizationMechanism(int)
     */
    public final static int AUTHORIZATION_MECHANISM_JAAS = 0;

    /**
     * Constant representing custom authorization
     *
     * @see ObjectGrid#setAuthorizationMechanism(int)
     */
    public final static int AUTHORIZATION_MECHANISM_CUSTOM = 1;

    /**
     * Constant indicating TCP/IP is the only supported transport.
     * <p>
     * If the client's transport type is set to this value,
     * TCP/IP is the only supported transport type. If the server
     * requires SSL, the client won't be able to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setTransportType(int)
     */
    public static final int TCP_IP =20;

    /**
     * Constant indicating SSL transport is supported.
     * <p>
     * If the client's transport type is set to this value,
     * the client supports both TCP/IP and SSL. SSL will be used
     * if both sides side supports SSL. Otherwise, TCP/IP will be used.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setTransportType(int)
     */
    public static final int SSL_SUPPORTED = 21;

    /**
     * Constant indicating SSL transport is required.
     * <p>
     * If the client's transport type is set to this value,
     * SSL is the only supported transport type. If the server
     * requires TCP/IP, the client won't be able to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setTransportType(int)
     */
    public static final int SSL_REQUIRED = 22;

    /**
     * Constant indicating credential authentication is not supported.
     * <p>
     * If the credential authentication type is set to this value, no credential
     * authentication will be enforced. If the server requires credential
     * authentication, the client won't be able to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setCredentialAuthenticationType(int)
     */
    public static final int CREDENTIAL_AUTHENTICATION_NEVER=40;

    /**
     * Constant indicating credential authentication is supported.
     * <p>
     * If the credential authentication type is set to this value, credential
     * authentication will be enforced if and only if both client and server
     * support credential authentication.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setCredentialAuthenticationType(int)
     */
    public static final int CREDENTIAL_AUTHENTICATION_SUPPORTED=41;

    /**
     * Constant indicating credential authentication is required.
     * <p>
     * If the credential authentication type is set to this value, credential
     * authentication will be enforced. If the server doesn't support
     * credential authentication, the client won't be able to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setCredentialAuthenticationType(int)
     */
    public static final int CREDENTIAL_AUTHENTICATION_REQUIRED=42;

    /**
     * Constant indicating client certificate authentication is not supported.
     * <p>
     * If the client certificate authentication type is set to this value, no
     * client certificate authentication will be enforced. If the server
     * doesn't support client certificate authentication, the client won't be able
     * to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setClientCertificateAuthentication(int)
     */
    public static final int CLIENT_CERTIFICATE_AUTHENTICATION_NEVER=50;

    /**
     * Constant indicating client certificate authentication is supported.
     * <p>
     * If the client certificate authentication type is set to this value,
     * client certificate authentication will be enforced when the following
     * conditions are met:
     * <ul>
     * <li> both client and server supports or requires client certificate authentication;
     * <li> the transport protocol to use is SSL;
     * <li> no credential authentication will be done.
     * </ul>
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setClientCertificateAuthentication(int)
     */
    public static final int CLIENT_CERTIFICATE_AUTHENTICATION_SUPPORTED=51;

    /**
     * Constant indicating client certificate authentication is required.
     *
     * If the client certificate authentication type is set to this value,
     * client certificate authentication will be enforced if the following
     * conditions are met:
     * <ul>
     * <li> both client and server supports or requires client certificate authentication;
     * <li> the transport protocol to use is SSL;
     * <li> no credential authentication will be done.
     * </ul>
     *
     * If the server doesn't support client certificate authentication
     * and no credential authentication will be done, the client won't be able
     * to connect to the server.
     *
     * @since WAS XD 6.0.1
     *
     * @see ClientSecurityConfiguration#setClientCertificateAuthentication(int)
     */
    public static final int CLIENT_CERTIFICATE_AUTHENTICATION_REQUIRED=52;

    /**
     * String representation for value Never indicating an option is not supported.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration settings
     * "clientCertificateAuthentication" and "credentialAuthentication".
     *
     * @since WAS XD 6.0.1
     *
     * @see #CLIENT_CERTIFICATE_AUTHENTICATION_NEVER
     * @see #CREDENTIAL_AUTHENTICATION_NEVER
     */
    public static final String NEVER_STRING = "Never";

    /**
     * String representation for value Supported indicating an option is supported.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration settings
     * "clientCertificateAuthentication" and "credentialAuthentication".
     *
     * @since WAS XD 6.0.1
     *
     * @see #CLIENT_CERTIFICATE_AUTHENTICATION_SUPPORTED
     * @see #CREDENTIAL_AUTHENTICATION_SUPPORTED
     */
    public static final String SUPPORTED_STRING = "Supported";

    /**
     * String representation for value Required indicating an option is required.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration settings
     * "clientCertificateAuthentication" and "credentialAuthentication".
     *
     * @since WAS XD 6.0.1
     *
     * @see #CLIENT_CERTIFICATE_AUTHENTICATION_REQUIRED
     * @see #CREDENTIAL_AUTHENTICATION_REQUIRED
     */
    public static final String REQUIRED_STRING = "Required";

    /**
     * String representation for value TCP/IP indicating a transport type of
     * TCP/IP is used.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration setting
     * "transportType".
     *
     * @since WAS XD 6.0.1
     *
     * @see #TCP_IP
     */
    public static final String TCPIP_STRING = "TCP/IP";

    /**
     * String representation for value SSL-Supported indicating SSL transport
     * type is supported.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration setting
     * "transportType".
     *
     * @since WAS XD 6.0.1
     *
     * @see #SSL_SUPPORTED
     */
    public static final String SSL_SUPPORTED_STRING = "SSL-Supported";

    /**
     * String representation for value SSL-Required indicating SSL transport type
     * is required.
     * <p>
     * This value is used as a value to configuration settings in a
     * <code>Properties</code> object or property file for client and server
     * security configurations. It is used for the configuration setting
     * "transportType".
     *
     *
     * @since WAS XD 6.0.1
     *
     * @see #SSL_REQUIRED
     */
    public static final String SSL_REQUIRED_STRING = "SSL-Required";

    /**
     * String representation for "none" type of the secure token manager.
     * <p>
     * This value is used in a property file for server security
     * configurations. It is used for the configuration setting
     * "secureTokenManagerType".
     *
     * @since WAS XD 6.0.1
     */
    public static final String SECURE_TOKEN_MANAGER_NONE_STRING = "none";

    /**
     * String representation for "default" type of the secure token manager.
     * <p>
     * This value is used in a property file for server security
     * configurations.  It is used for the configuration setting
     * "secureTokenManagerType". This value requires users to provide the secure
     * token key store settings.
     *
     * @since WAS XD 6.0.1
     */
    public static final String SECURE_TOKEN_MANAGER_DEFAULT_STRING = "default";

    /**
     * String representation for "custom" type of the secure token manager.
     * <p>
     * This value is used in a property file for server security
     * configurations. It is used for the configuration setting
     * "secureTokenManagerType". This value requires users to provide the
     * SecureTokenManager implementation class name using the
     * "customSecureTokenManagerClass" configuration setting.
     *
     * @since WAS XD 6.0.1
     */
    public static final String SECURE_TOKEN_MANAGER_CUSTOM_STRING = "custom";
    
    /**
     * String representation for "autoSecret" type of the secure token manager.
     * <p>
     * This value is used in a property file for server security
     * configurations. It is used for the configuration setting
     * "secureTokenManagerType". This value does not require users to provide
     * other settings.
     *
     * @since 8.6
     */
    public static final String NEW_SECURE_TOKEN_MANAGER_STRING = "autoSecret";

    /**
     * The access by creator only authorization is disabled.
     * <p>
     * The access by creator authorization ensures that only the user (represented
     * by the Principals associated with it), who inserts the data entry into the map,
     * can access the data. Here the access means read, update, invalidate, and
     * remove.
     * @since WAS XD 6.1 FIX3
     */
    public static final int ACCESS_BY_CREATOR_ONLY_DISABLED = 60;

    /**
     * The access by creator only authorization is enabled to complement the
     * ObjectGrid map authorization.
     * <p>
     * The access by creator authorization ensures that only the user (represented
     * by the Principals associated with it), who inserts the data entry into the map,
     * can access the data. Here the access means read, update, invalidate, and
     * remove.
     * <p>
     * If this constant is used, both map authorization and access by creator only
     * authorization will take effect. Therefore, you can further limit
     * the operations to the data entries. For example, you can restrict the creator
     * from invalidating the data entries.
     * @since WAS XD 6.1 FIX3
     */
    public static final int ACCESS_BY_CREATOR_ONLY_COMPLEMENT = 61;

    /**
     * The access by creator only authorization is enabled to supersede the
     * ObjectGrid map authorization.
     * <p>
     * The access by creator authorization ensures that only the user (represented
     * by the Principals associated with it), who inserts the data entry into the map,
     * can access the data. Here the access means read, update, invalidate, and
     * remove.
     * <p>
     * If this constant is used, the access by creator only authorization
     * will supersede the map authorization; no map authorization will be
     * done.
     * @since WAS XD 6.1 FIX3
     */
    public static final int ACCESS_BY_CREATOR_ONLY_SUPERSEDE = 62;
}