// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.2 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/security/plugins/Credential.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:24:22
//
// Module  :  Credential.java
//
// Source File Description: 
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD601_M3_Bx  WASX.XD   11/15/2005   jitang    ObjectGrid security support
// PK17543      WASX.XD   01/11/2006   jitang    Update JavaDoc
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security.plugins;

import java.io.Serializable;

/**
 * <p>This interface represents a credential used by an ObjectGrid client. An instance 
 * of this class represents one client indentity. This credential will be sent to the 
 * ObjectGrid server for authentication. It should be serializable.
 * 
 * <p>A credential has to implement the <code>equals(Object)</code> and 
 * <code>hashCode()</code> methods.  Two Credential objects are considered equal 
 * if and only if they represent the same identity and security information. For 
 * example, if the credential contains a user ID and password. Two credentials 
 * are equal if and only if both their user IDs and passwords are equal.
 * 
 * <p>ObjectGrid provides three built-in implementations for this interface:
 * <ul> 
 * <li>com.ibm.websphere.objectgrid.security.plugins.builtins.ClientCertificateCredential:
 * A credential containing an SSL certificate chain. </li>
 * <li>com.ibm.websphere.objectgrid.security.plugins.builtins.UserPasswordCredential:
 * A credential containing a user ID and password pair. </li>
 * <li>com.ibm.websphere.objectgrid.security.plugins.builtins.WSTokenCredential:
 * A credential containing WebSphere Application Server specific authentication 
 * and authorization tokens. </li>
 * </ul>
 *  
 * <P>Refer to the respective JavaDoc for more details.</p>
 * 
 * @since WAS XD 6.0.1, XC10
 * 
 * @see CredentialGenerator
 */
public interface Credential extends Serializable {

    /**
     * Checks two Credential objects for equality.
     * <p>
     * Two Credential objects are considered equal if and only if
     * they represent the same identity and security information.
     * 
     * @param o the object we are testing for equality with this object.
     *
     * @return <code>true</code> if both Credential objects are equivalent.
     */
    boolean equals(Object o);

    /**
     * Returns the hashcode of the Credential object
     * 
     * @return the hash code of the Credential object
     */
    int hashCode();
}

